/*****
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****/


#include "Config.h"
#include "net/unicode.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <map>
#include "iconv.h"
#include "string_utils.h"

// MS Language Strings
// - http://msdn.microsoft.com/library/default.asp?url=/library/en-us/vclib/html/_crt_language_strings.asp
static const char *codepage[] = 
{
	"ARA", "CP1256", "ar_SA",  // Arabic (Saudi Arabia)
	"CAT", "CP1252", "ca_ES",  // Catalan    
	"CHT", "CP950",  "zh_TW",  // Chinese (Taiwan)
	"CHS", "CP936",  "zh_CN",  // Chinese (People's Republic of China)
	"CSY", "CP1250", "cs_CZ",  // Czech (Czech Republic)
	"DAN", "CP1252", "da_DK",  // Danish (Denmark)
	"DEU", "CP1252", "de_DE",  // German (Germany)
	"ELL", "CP1253", "el_GR",  // Greek (Greece)
	"ENU", "CP1252", "en_US",  // English (United States)
	"ESN", "CP1252", "es_ES",  // Spanish - Modern Sort (Spain)
	"ESP", "CP1252", "es_ES",  // Spanish    
	"ETI", "CP1257", "et_EE",  // Estonian   
	"FIN", "CP1252", "fi_FI",  // Finnish (Finland)
	"FRA", "CP1252", "fr_FR",  // French (France)
	"HEB", "CP1255", "he_IL",  // Hebrew (Israel)
	"HUN", "CP1250", "hu_HU",  // Hungarian (Hungary)
	"ITA", "CP1252", "it_IT",  // Italian (Italy)
	"JPN", "CP932",  "ja_JP",  // Japanese (Japan)
	"KOR", "CP949",  "ko_KR",  // Korean (Korea)
	"LTH", "CP1257", "lt_LT",  // Lithuanian 
	"LVI", "CP1257", "lv_LV",  // Latvian    
	"NLD", "CP1252", "nl_NL",  // Dutch (Netherlands)
	"NOR", "CP1252", "no_NO",  // Norwegian (Bokmal) (Norway)
	"PLK", "CP1250", "pl_PL",  // Polish (Poland)
	"PTB", "CP1252", "pt_BR",  // Portuguese (Brazil)
	"PTG", "CP1252", "pt_PT",  // Portuguese (Portugal)
	"ROM", "CP1250", "ro_RO",  // Romanian (Romania)
	"RUS", "CP1251", "ru_RU",  // Russian (Russia)
	"HRV", "CP1250", "hr_HR",  // Croatian (Croatia)
	"SKY", "CP1250", "sk_SK",  // Slovak (Slovakia)
	"SVE", "CP1252", "sv_SE",  // Swedish (Sweden)
	"THA", "CP874",  "th_TH",  // Thai (Thailand)
	"TRK", "CP1254", "tr_TR",  // Turkish (Turkey)
	"SLV", "CP1250", "sl_SI",  // Slovenian (Slovenia)
	"VIT", "CP1258", "vi_VN",  // Vietnamese (Viet Nam)
	"EUQ", "CP1252", "eu_ES",  // Basque     
	NULL, NULL, NULL
};

class cLangMap : public std::map<std::string, int> 
{
public:
	cLangMap()
	{
		int i = 0;
		while (codepage[i*3])
		{
			insert(std::make_pair(codepage[i*3], i));
			i++;
		}
	}

	const char *getLocaleLang(const char *lang)
	{
		if (strlen(lang) > 0)
			return lang;

		if (!nConfig::locale_lang.empty())
			return nConfig::locale_lang.c_str();

		iterator it = find(stringUpper(nConfig::cliloc_lang));
		if (it != end())
			nConfig::locale_lang = codepage[it->second * 3 + 2];
		else
			nConfig::locale_lang = "us_EN"; // English

		return nConfig::locale_lang.c_str();

	}

	const char *getCodePage(const char *charset)
	{
		if (strlen(charset) > 0) 
			return charset;
		
		if (!nConfig::charset.empty()) 
			return nConfig::charset.c_str();
		
		iterator it = find(stringUpper(nConfig::cliloc_lang));
		if (it != end())
			nConfig::charset = codepage[it->second * 3 + 1];
		else
			nConfig::charset = "CP1252"; // ASCII

		return nConfig::charset.c_str();
	}
} mapLang;

const char *getLocaleLang(const char *locale_lang)
{
	return mapLang.getLocaleLang(locale_lang);
}

cUnicode::cUnicode (const char *str, int len /* = 0 */, bool network /*= true*/, const char *charset /*= ""*/ )
{
	memset (m_charBuf, 0, sizeof (m_charBuf));
	memset (m_unicodeBuf, 0, sizeof (m_unicodeBuf));

	if (len <= 0)
		len = strlen(str);

	if (len > sizeof(m_charBuf) - 1)
		len = sizeof(m_charBuf) - 1;

	strncpy (m_charBuf, str, len);

	m_strLen = 0;
	m_charLen = len + 1; // +1 : '\0'

	CharToUnicode (mapLang.getCodePage(charset), network);
}

cUnicode::cUnicode (const UNICHAR *str, int len /* = 0 */, bool network /*= true*/, const char *charset /*= ""*/ )
{
	memset (m_charBuf, 0, sizeof (m_charBuf));
	memset (m_unicodeBuf, 0, sizeof (m_unicodeBuf));

	if (len <= 0 || len > sizeof(m_charBuf) - 1) 
		len = sizeof(m_charBuf) - 1;

	UnicodeCpy (str, len);

	UnicodeToChar (mapLang.getCodePage(charset), network);
}

cUnicode::~cUnicode ()
{
}

void cUnicode::UnicodeCpy (const UNICHAR * str, int maxlen)
{
	UNICHAR *n = m_unicodeBuf;

	while (*str && maxlen)
	{
		*(n++) = *(str++);
		maxlen--;
	}

	*n = 0;

	m_strLen = 0;
	m_unicodeLen = (n - m_unicodeBuf + 1) * sizeof(UNICHAR);
}

void cUnicode::CharToUnicode(const char *charset, bool network)
{
#ifdef WIN32
	const char *in  = m_charBuf;
#else
	char *in  = m_charBuf;
#endif

	char *out = (char*)m_unicodeBuf;
	size_t inLen  = m_charLen;
	size_t outLen = sizeof(m_unicodeBuf);

	iconv_t cd = iconv_open(network ? "UCS-2" : "UCS-2-INTERNAL", charset);
	if (cd < (iconv_t)0) return;
	
	if (iconv(cd, &in, &inLen, &out, &outLen) == (size_t)-1)
	{
		// converting error
		m_strLen = 0;
	}
	else
	{
		m_strLen = (UNICHAR*)out - m_unicodeBuf - 1;  // -1 '\0'
		m_unicodeLen = (m_strLen + 1) * sizeof(UNICHAR);
	}

	iconv_close(cd);
}

void cUnicode::UnicodeToChar(const char *charset, bool network)
{
#ifdef WIN32
	const char *in  = (char*)m_unicodeBuf;
#else
	char *in  = (char*)m_unicodeBuf;
#endif

	char *out = m_charBuf;
	size_t inLen  = m_unicodeLen;
	size_t outLen = sizeof(m_charBuf);

	iconv_t cd = iconv_open(charset, network ? "UCS-2" : "UCS-2-INTERNAL");
	if (cd < (iconv_t)0) return;
	
	if (iconv(cd, &in, &inLen, &out, &outLen) == (size_t)-1)
	{
		// converting error
		m_strLen = 0;
	}
	else
	{
		m_strLen  = out - m_charBuf - 1;  // -1 '\0'
		m_charLen = m_strLen + 1;
	}

	iconv_close(cd);
}


